@@ -123,16 +123,17 @@ class OPTemplate:
                         value="{{pod.name}}",
                         global_name="{{inputs.parameters.dflow_key}}",
                     )
-                    self.outputs.parameters["dflow_wfname"] = OutputParameter(
-                        value="{{workflow.name}}",
-                        global_name="dflow_workflow_name",
-                    )
                 else:
                     self.outputs.parameters["dflow_global"] = OutputParameter(
                         value_from_parameter="non-exists",
                         default="{{node.id}}",
                         global_name="{{inputs.parameters.dflow_key}}",
                     )
+        elif "dflow_group_key" in self.inputs.parameters:
+            if memoize_prefix is not None:
+                self.memoize_key = \
+                    "%s-{{inputs.parameters.dflow_group_key}}-init-artifact" \
+                    % memoize_prefix
 
         if self.memoize_key is not None:
             # Is it a bug of Argo?
