@@ -1,6 +1,7 @@
 import json
 import logging
 import os
+import time
 from copy import deepcopy
 from typing import Any, Dict, List, Optional, Union
 
@@ -319,6 +320,97 @@ class Workflow:
                                                                  self.uid))
         return workflow
 
+    def wait(self, interval=1):
+        while self.query_status() in ["Pending", "Running"]:
+            time.sleep(interval)
+
+    def handle_reused_step(self, step):
+        outputs = {}
+        if hasattr(step, "outputs"):
+            if hasattr(step.outputs, "exitCode"):
+                outputs["exitCode"] = step.outputs.exitCode
+            if hasattr(step.outputs, "parameters"):
+                outputs["parameters"] = []
+                for name, par in step.outputs.parameters.items():
+                    if not hasattr(step.outputs.parameters[name],
+                                   "save_as_artifact"):
+                        outputs["parameters"].append(par.recover())
+            if hasattr(step.outputs, "artifacts"):
+                for name, art in step.outputs.artifacts.items():
+                    if hasattr(step, "inputs") and \
+                        hasattr(step.inputs, "parameters") and \
+                        "dflow_group_key" in step.inputs.parameters \
+                            and name != "main-logs":
+                        if config["overwrite_reused_artifact"]:
+                            self.handle_reused_artifact(step, name, art)
+                        else:
+                            self.handle_reused_artifact_with_copy(
+                                step, name, art)
+
+        outputs["artifacts"] = [
+            art.recover() for art in step.outputs.artifacts.values()]
+        self.memoize_map["%s-%s" % (self.id, step.key)] = {
+            "nodeID": step.id,
+            "outputs": outputs,
+            "creationTimestamp": step.finishedAt,
+            "lastHitTimestamp": step.finishedAt
+        }
+
+    def handle_reused_artifact(self, step, name, art):
+        group_key = step.inputs.parameters["dflow_group_key"].value
+        wf_name = step.workflow
+        memoize_key = "%s-%s-init-artifact" % (self.id, group_key)
+        if memoize_key not in self.memoize_map:
+            self.memoize_map[memoize_key] = {
+                "nodeID": "dflow-%s" % randstr(10),
+                "outputs": {
+                    "parameters": [{
+                        "name": "dflow_artifact_key",
+                        "value": "%s/%s" % (wf_name, group_key),
+                    }],
+                    "artifacts": [],
+                },
+                "creationTimestamp": step.finishedAt,
+                "lastHitTimestamp": step.finishedAt
+            }
+        init_step = self.memoize_map[memoize_key]
+        art_key = get_key(art, raise_error=False)
+        if group_key not in art_key:
+            key = "%s%s/%s/%s" % (s3_config["prefix"], wf_name, group_key,
+                                  name)
+            logger.debug("copying artifact: %s -> %s" % (art_key, key))
+            copy_s3(art_key, key)
+            art_key = key
+
+        init_art = None
+        for art in init_step["outputs"]["artifacts"]:
+            if art["name"] == name:
+                init_art = art
+                break
+        if not init_art:
+            init_art = {"name": name}
+            if s3_config["repo_type"] == "s3":
+                init_art["s3"] = {
+                    "key": art_key,
+                    "archive": {"none": {}},
+                }
+            elif s3_config["repo_type"] == "oss":
+                init_art["oss"] = {
+                    "key": art_key,
+                    "archive": {"none": {}},
+                }
+            init_step["outputs"]["artifacts"].append(init_art)
+
+    def handle_reused_artifact_with_copy(self, step, name, art):
+        old_key = get_key(art, raise_error=False)
+        if old_key and old_key not in self.copied_keys:
+            key = "%s%s/%s/%s" % (
+                s3_config["prefix"], self.id, step.inputs.parameters[
+                    "dflow_group_key"].value, name)
+            logger.debug("copying artifact: %s -> %s" % (old_key, key))
+            copy_s3(old_key, key)
+            self.copied_keys.append(old_key)
+
     def convert_to_argo(self, reuse_step=None):
         if self.context is not None:
             assert isinstance(self.context, (Context, Executor))
@@ -327,64 +419,34 @@ class Workflow:
         status = None
         if reuse_step is not None:
             self.id = self.name + "-" + randstr()
-            copied_keys = []
-            reused_keys = []
+            self.copied_keys = []
+            self.memoize_map = {}
+            key2id = {}
             for step in reuse_step:
                 data = {}
                 if step.key is None:
                     continue
-                reused_keys.append(step.key)
-                outputs = {}
-                if hasattr(step, "outputs"):
-                    if hasattr(step.outputs, "exitCode"):
-                        outputs["exitCode"] = step.outputs.exitCode
-                    if hasattr(step.outputs, "parameters"):
-                        outputs["parameters"] = []
-                        for name, par in step.outputs.parameters.items():
-                            if not hasattr(step.outputs.parameters[name],
-                                           "save_as_artifact"):
-                                outputs["parameters"].append(par.recover())
-                    if hasattr(step.outputs, "artifacts"):
-                        for name, art in step.outputs.artifacts.items():
-                            if hasattr(step, "inputs") and \
-                                hasattr(step.inputs, "parameters") and \
-                                "dflow_group_key" in step.inputs.parameters \
-                                    and name != "main-logs":
-                                old_key = get_key(art, raise_error=False)
-                                if old_key and old_key not in copied_keys:
-                                    key = "%s%s/%s/%s" % (
-                                        s3_config["prefix"], self.id,
-                                        step.inputs.parameters[
-                                            "dflow_group_key"].value, name)
-                                    logger.debug("copying artifact: %s -> %s"
-                                                 % (old_key, key))
-                                    copy_s3(old_key, key)
-                                    copied_keys.append(old_key)
-                        outputs["artifacts"] = [
-                            art.recover()
-                            for art in step.outputs.artifacts.values()]
-                data["%s-%s" % (self.id, step.key)] = json.dumps({
-                    "nodeID": step.id,
-                    "outputs": outputs,
-                    "creationTimestamp": step.finishedAt,
-                    "lastHitTimestamp": step.finishedAt
-                })
+                key2id[step.key] = step.id
+                self.handle_reused_step(step)
+
+            for key, step in self.memoize_map.items():
+                data = {key: json.dumps(step)}
                 config_map = kubernetes.client.V1ConfigMap(
                     data=data, metadata=kubernetes.client.V1ObjectMeta(
-                        name="dflow-%s-%s" % (self.id, step.key)))
+                        name="dflow-%s" % key))
                 core_v1_api = self.get_k8s_core_v1_api()
-                logger.debug("creating configmap: dflow-%s-%s" %
-                             (self.id, step.key))
+                logger.debug("creating configmap: %s" %
+                             config_map.metadata.name)
                 core_v1_api.api_client.call_api(
                     '/api/v1/namespaces/%s/configmaps' % self.namespace,
                     'POST', body=config_map, response_type='V1ConfigMap',
                     header_params=config["http_headers"],
                     _return_http_data_only=True)
-            self.handle_template(
-                self.entrypoint, memoize_prefix=self.id,
-                memoize_configmap="dflow")
-            status = {"outputs": {"parameters": [{"name": key} for key in
-                                                 reused_keys]}}
+
+            self.handle_template(self.entrypoint, memoize_prefix=self.id,
+                                 memoize_configmap="dflow")
+            status = {"outputs": {"parameters": [
+                {"name": key, "value": id} for key, id in key2id.items()]}}
         else:
             self.handle_template(self.entrypoint)
 
@@ -666,8 +728,7 @@ class Workflow:
                     header_params=config["http_headers"],
                     query_params=[('fields', 'status.outputs')])
             return [par["name"] for par in
-                    response["status"]["outputs"]["parameters"]
-                    if par["name"] != "dflow_workflow_name"]
+                    response["status"]["outputs"]["parameters"]]
         except Exception:
             logger.warn("Key-ID map not found in the global outputs, downgrade"
                         " to full query")
@@ -689,25 +750,22 @@ class Workflow:
         if isinstance(key, str):
             key = [key]
 
-        workflow = self.query(fields=['status.outputs'])
         try:
-            parameters = workflow.status.outputs.parameters
-            wf_name = next(filter(lambda par: par["name"] ==
-                                  "dflow_workflow_name", parameters))["value"]
+            workflow = self.query(fields=['metadata.name', 'status.outputs'])
+            wf_name = workflow.metadata.name
+            key2id = {}
+            for par in workflow.status.outputs.parameters:
+                pod_name = par["value"]
+                key2id[par["name"]] = wf_name + "-" + pod_name.split("-")[-1]
+
+            workflow = self.query(
+                fields=['metadata.name'] + [
+                    'status.nodes.' + key2id[k] for k in key])
+            return workflow.get_step()
         except Exception:
             logger.warn("Key-ID map not found in the global outputs, downgrade"
                         " to full query")
             return self.query_step(key=key)
-        key2id = {}
-        for par in workflow.status.outputs.parameters:
-            if par["name"] == "dflow_workflow_name":
-                continue
-            pod_name = par["value"]
-            key2id[par["name"]] = wf_name + "-" + pod_name.split("-")[-1]
-
-        workflow = self.query(
-            fields=['status.nodes.' + key2id[k] for k in key])
-        return workflow.get_step()
 
     def query_global_outputs(self) -> ArgoWorkflow:
         """
@@ -719,7 +777,7 @@ class Workflow:
         Returns:
             a list of steps
         """
-        workflow = self.query(fields=['status.outputs'])
+        workflow = self.query(fields=['metadata.name', 'status.outputs'])
         step = ArgoStep(workflow.status)
         if hasattr(step, "outputs"):
             return step.outputs
